;;-----------------------------LICENSE NOTICE------------------------------------
;; This file is part of Invasion a Marte: An Amstrad CPC Game 
;; Copyright (C) 2019 
;; Authors:
;;  * Antonio Macia Lillo       tonetml23@gmail.com
;;  * Gina Esthela Abad Jimenez abadjimenezgina@gmail.com
;;  * Ivan Martinez Molina      imm79@alu.ua.es
;;
;;  This program is free software: you can redistribute it and/or modify
;;  it under the terms of the GNU Lesser General Public License as published by
;;  the Free Software Foundation, either version 3 of the License, or
;;  (at your option) any later version.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU Lesser General Public License for more details.
;;
;;  You should have received a copy of the GNU Lesser General Public License
;;  along with this program.  If not, see <http://www.gnu.org/licenses/>.
;;-------------------------------------------------------------------------------

;;
;; ENTITY MANAGER
;;
.include "man/entity_manager.h.s"
.include "cpctelera.h.s"
.include "man/collision_manager.h.s"
.include "sys/render_system.h.s"

.module entity_manager

DefineComponentArrayStructure _entity, max_entities, DefaultEntity
_last_elem_ptr: .dw _entity_array

;;Arrays of pointers
DefineComponentArrayStructure_Size _collision, max_entities, 2
DefineComponentArrayStructure_Size _delete, max_entities, 2

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Gets the entity array pointer on IX and
;; the number of entities on A
;;
;;INPUT: -
;;DESTROYS: 
;; IX, A
;;
entityman_getEntityArray::
    ld  ix, #_entity_array
    ld  a,  (_entity_num)
    ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Gets the collision array pointer on IX and
;; the number of pointers on A
;;
;;INPUT: -
;;DESTROYS: 
;; IX, A
;;
entityman_getCollisionArray::
    ld  ix, #_collision_array
    ld  a,  (_collision_num)
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Gets the delete array pointer on IX and
;; the number of pointers on A
;;
;;INPUT: -
;;DESTROYS: 
;; IX, A
;;
entityman_getDeleteArray::
    ld  ix, #_delete_array
    ld  a,  (_delete_num)
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Initializes the entity manager. It sets
;; up everything with 0 entities and ready 
;; to start creating new ones
;;
;;INPUT: -
;;DESTROYS: AF, HL
;;

;;  INPUT
;;  HL: Pointer to entity initializer bytes
;;  
entityman_init::
    ;;Reset all entity component vector values
    xor a
    ld (_entity_num), a

    ld hl, #_entity_array
    ld (_entity_pend), hl 

    ;;Reset all collision component vector values
    xor a
    ld (_collision_num), a

    ld hl, #_collision_array
    ld (_collision_pend), hl

    call entityman_reset_deletearr

    ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Updates the entity manager
;;
;;INPUT: -
;;DESTROYS: AF, HL
;;

;;  INPUT -
;;  
entityman_update::
    ;; Delete entities on the delete array
    ld a, (_delete_num)
    ld ix, #_delete_array                       ;; Load delete array pointer on ix

_del_arr_loop:
    or a
    jp z, _del_end_loop                        ;; While a > 0 loop

    ;;Load entity pointer on hl
    ld l, (ix)
    inc ix
    ld h, (ix)
    inc ix

    push af
    push ix
    call entityman_delete                       ;; Delete entity pointed by hl
    pop ix
    pop af
    dec a
    jp _del_arr_loop
_del_end_loop:

    ;;Reset delete array
    call entityman_reset_deletearr
    ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Resets the delete array
;;
;;INPUT: -
;;DESTROYS: AF, HL
;;  
entityman_reset_deletearr::
    ;;Reset all delete component vector values
    xor a
    ld (_delete_num), a

    ld hl, #_delete_array
    ld (_delete_pend), hl 

    ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Inserts entity pointer into delete array
;;
;;INPUT:
;; DE: Pointer to the entity
;;DESTROYS: HL, BC, AF, DE, IX
;;RETURN:
;; BC: Sizeof (Entity_T)
;;
entityman_deletearr_add::
    ;; Check if has already been added
    ld ix, #_delete_array
    ld b, d                 ;; Load pointer to search on bc
    ld c, e
    ld a, (_delete_num)
    push de
    call entityman_exists
    pop de
    ret z

    ;; Increment number of reserved delete pointers
    ld hl, #_delete_num
    inc (hl)

    ;;Increment Array end pointer to point to the next
    ;;free element in the array
    ld hl, (_delete_pend)

    ;;Save pointer in the array
    ld (hl), e
    inc hl
    ld (hl), d
    inc hl

    ;;Save array end pointer
    ld (_delete_pend), hl

    ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Initializes a new entity with 
;; empty values
;;
;;INPUT: -
;;DESTROYS: DE, HL, BC
;;RETURN:
;; DE: Points to added element
;; BC: Sizeof (Entity_T)
;;
entityman_new::
    ;; Increment number of reserved entities
    ld hl, #_entity_num
    inc (hl) ;;entity_num++

    ;;Increment Array end pointer to point to the next
    ;;free element in the array
    ld hl, (_entity_pend)
    ld d, h
    ld e, l
    ld bc, #sizeof_e
    add hl, bc
    ld (_entity_pend), hl

    ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Initializes a new entity with 
;; empty values
;;
;;INPUT: 
;; IX: pointers array
;; A: number of entities
;; BC: entity to search
;;DESTROYS: DE, HL, BC, IX
;;RETURN:
;; Z set if exists
;;
entityman_exists::
_ea_loop:
    or a
    jp z, _ea_end_loop

    ld h, 1(ix)
    ld l, 0(ix)

    or a                ;; Reset carry flag
    sbc hl, bc
    jp z, _ea_exists

    ld de, #sizeof_d
    add ix, de

    dec a
    jp _ea_loop

_ea_end_loop:
    ld a, #1
    or a                ;; Reset Z flag
    ret

_ea_exists:
    xor a
    ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Inserts entity pointer into collision array
;;
;;INPUT:
;; DE: Pointer to the entity
;;DESTROYS: HL, BC
;;RETURN:
;; BC: Sizeof (Entity_T)
;;
entityman_collision_add::
    ;; Increment number of reserved collision pointers
    ld hl, #_collision_num
    inc (hl)

    ;;Increment Array end pointer to point to the next
    ;;free element in the array
    ld hl, (_collision_pend)

    ;;Save pointer in the array
    ld (hl), e
    inc hl
    ld (hl), d
    inc hl

    ;;Save array end pointer
    ld (_collision_pend), hl

    ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Creates a new entity whit its values
;;
;;INPUT:
;; HL: Pointer to entity initializer bytes
;;DESTROYS: DE, HL, BC, IX
;;
entityman_create::
    push hl
    call entityman_new
    
    ;; IX = DE
    ld__ixh_d
    ld__ixl_e
    ;;Copy initialization values to new entity
    ;;HL entity values
    ;;DE points to the new added entity
    ;;BC holds sizeof(Entity_t)
    ;;Copy initialization values
    ld bc, #sizeof_e
    pop hl
    push de
    ldir
    pop de

    ;;ADD COLLISION HANDLERS
    ld a, e_collider(ix)
    cp #coll_type_nocoll
    call nz, entityman_collision_add    ;; Add to collision array if needed

    ;;Add default handler
    ld hl, #notihng_collides
    ld e_collision_handler_l(ix), l     ;; Save it to memory
    ld e_collision_handler_h(ix), h

    ;; Add player handler
    ld a, e_collider(ix)                ;; Load type of collider entity type to a
    cp #coll_type_player
    jp nz, _not_player_collider         ;; IF collyder type is player, load the collider handler pointer
    ld hl, #player_collides              ;; Load handler memory adress to hl
    ld e_collision_handler_l(ix), l     ;; Save it to memory
    ld e_collision_handler_h(ix), h

    ;; Add bullet handler
_not_player_collider:
    ld a, e_collider(ix)                ;; Load type of collider entity type to a
    cp #coll_type_bullet
    jp nz, _not_bullet_collider         ;; IF collyder type is bullet, load the collider handler pointer
    ld hl, #bullet_collides              ;; Load handler memory adress to hl
    ld e_collision_handler_l(ix), l     ;; Save it to memory
    ld e_collision_handler_h(ix), h

    ;; Add enemy handler
_not_bullet_collider:
    ld a, e_collider(ix)                ;; Load type of collider entity type to a
    cp #coll_type_enemy
    jp nz, _not_enemy_collider         ;; IF collyder type is bullet, load the collider handler pointer
    ld hl, #enemy_collides              ;; Load handler memory adress to hl
    ld e_collision_handler_l(ix), l     ;; Save it to memory
    ld e_collision_handler_h(ix), h

_not_enemy_collider:
    ret


ead_ent_pointer: .dw 0x0000
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Gets the number of entities left from that possition on the array on a
;;
;;INPUT:
;; HL: Pointer to entity
;;
;;DESTROYS: IX, AF, DE, HL
;;
;;RETURNS:
;; A: Entities left
;;
entityman_entities_left::
    xor a                           ;; Set a to 0
    ld bc, #sizeof_e
    ld (ead_ent_pointer), hl

_eal_loop:

    ld de, (_entity_pend)

    ld hl, (ead_ent_pointer)
    add hl, bc                      ;; Increment entity pointer
    ld (ead_ent_pointer), hl        ;; Store it in memory
    inc a

    or a                            ;; Reset the carry flag
    sbc hl, de                      ;; If ent ptr < end ptr a++
    jp c, _eal_loop                 ;; When the entity pointer is mayor than the end of the array its time to stop
    ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Deletes the entity pointed in IX from the entity
;; array, and rearranges the array to compact it
;;
;;INPUT:
;; HL: Pointer to entity
;;DESTROYS: DE, HL, BC, IX, IY, AF
;;
entityman_delete::
    ;;Draw background square on its position
    push hl
    push hl
    pop ix                                  ;; Get entity pointer on ix
    call rendersys_draw_sprite_background   ;; Draw the background

    pop hl
    push hl
    ld d, h                             ;; Copy the pointer onto DE
    ld e, l

    push hl
    push de
    call entityman_entities_left        ;; Get number of iterations on a
    pop de
    pop hl

    ;;Make hl point to the next entity
    ld bc, #sizeof_e                    ;; BC points to the size of entity
    add hl, bc                          ;; HL points to the next object

_delete_loop:
    ld bc, #sizeof_e                    ;; Restore entity size as its deleted by ldir

    ;;Rewrite all possitions on the array to the left
    ;;HL pointer to the next entity
    ;;DE points to the entity to delete, wich is going to be overwritten
    ;;BC holds sizeof(Entity_t)
    dec a                               ;; Decrease entity loop counter
    ldir
    jp nz, _delete_loop

    ;;Decrease _entity_num
    ld a, (_entity_num)                 ;; Load number of entities
    dec a                               ;; Decrease number of entities by 1
    ld (_entity_num), a                 ;; Save the number

    ;;Decrease _entity_pend
    ld hl, (_entity_pend)               ;; Load pointer to the end of the array
    ld bc, #sizeof_e                    ;; Load size of entity on bc
    or a                                ;; Set carry flag to 0 so sbc wont substract it
    sbc hl, bc                          ;; End pointer - size of entity
    ld (_entity_pend), hl               ;; Save it onto memory

    ; ;; Rearrange arrays of pointers
    pop hl                              ;; IX: Entity deleted
    push hl

    ld a, (_collision_num)                ;; A: Number of pointers on the array to rearrange
    ld ix, #_collision_array             ;; IY: Array to rearrange
    ld de, #_collision_num              ;; Pointer to the number of collision elements
    ld bc, #_collision_pend             ;; Pointer to the end pointer
    call entityman_arraypointer_delete  ;; Delete the pointer of the object on the collision array

    pop ix
    push ix

    ld a, (_collision_num)                ;; A: Number of pointers on the array to rearrange
    ld iy, #_collision_array             ;; IY: Array to rearrange
    call entityman_rearrange

    pop ix                              ;; IX: Entity deleted
    ld a, (_delete_num)                ;; A: Number of pointers on the array to rearrange
    ld iy, #_delete_array             ;; IY: Array to rearrange
    call entityman_rearrange

    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Changes pointers of the array passed on IY to correct the error when an 
;; entity gets deleted, the position of the next entities on memory changes
;;
;;INPUT:
;; IX: Pointer of the entity deleted
;; IY: Array of pointers
;; A: Number of elements in the array of pointers
;;DESTROYS: HL, BC, AF, IX, IY
;;
entityman_rearrange::
    ld h, 1(iy)                         ;; Load first pointer on hl
    ld l, 0(iy)                         ;; (Little endian)
    ld bc, #sizeof_e                    ;; Element size

re_del_loop:
    push af
    ;;Check high possition
    ld__a_ixh
    cp h
    jp z, _siguiente
    jp nc, _no_mayor_delete             ;; IF ix < hl the pointer has not to be changed
    jp c, _doit                         ;; if ix > hl, there's no need of checking the low value

_siguiente:
    ;;Check low possition
    ld__a_ixl
    cp l
    jp nc, _no_mayor_delete             ;; IF ix < hl the pointer has not to be changed

_doit:
    or a                                ;; Set carry flag to 0 so its not substracted
    sbc hl, bc                          ;; all elements are one position back, so pointer - entity_size
    ld 1(iy), h                          ;; Save high part of the pointer on memory
    ld 0(iy), l                          ;; Save low part of the pointer on the memory

_no_mayor_delete:
    ;;Increment pointer of array
    inc iy
    inc iy
    ld h, 1(iy)
    ld l, 0(iy)
    pop af
    dec a                                ;; counter-- (number of pointers)
    jp nz, re_del_loop           

    ret

ead_local_counter: .db 0
ead_pointer_num: .db 0
ead_pointer_num_pointer: .dw 0x0000
ead_pointer_end_pointer: .dw 0x0000
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;DESCRIPTION
;; Deletes a pointer from the array
;; IX: Array of pointers
;; HL: Value of the pointer to delete
;; A: Number of pointers in the array
;;
;;INPUT:
;; HL: Pointer of the entity deleted
;; IX: Array of pointers
;; A: Number of elements in the array of pointers
;; DE: Pointer of the number of elements
;; BC: Pointer to the end pointer
;;DESTROYS: HL, BC, AF
;;
entityman_arraypointer_delete::
ld (ead_pointer_num_pointer), de
ld (ead_pointer_end_pointer), bc
ld (ead_pointer_num), a                 ;; Save the number of elements on memory
_ead_loop:
    ld (ead_local_counter), a           ;; Save counter on memory
    or a
    jp z, _end_ead_loop                 ;; If counter == 0 end     

    ;; Compare if the actual pointer is what we are looking for
    ;;Check high possition
    ld a, 1(ix)
    cp h
    jp nz, _ead_not_equals

    ;;Check low possition
    ld a, 0(ix)
    cp l
    jp nz, _ead_not_equals             ;; IF ix < hl the pointer has not to be changed
    
    push ix
    push hl
    ;;IF it is delete it
    ;;Rewrite all possitions on the array to the left
    ;;HL pointer to the next entity
    ;;DE points to the position to delete, wich is going to be overwritten
    ;;BC size of a pointer
    ld__d_ixh                           ;; De point to the actual pointer
    ld__e_ixl
    
    ld h, d
    ld l, e
    inc hl
    inc hl                              ;; Make hl point to the next entity
    
    ld b, #0
    ld a, (ead_local_counter)           ;; Number of entities left
    ld c, a
    rl c                                ;; num_pointers * 2 -> total size to move by ldir
    ldir

    ;;Decrease number of pointers
    ld a, (ead_pointer_num)            ;; Load the number of pointers in the array on a
    dec a                               ;; Decrease de number of pointers by one
    ld ix, (ead_pointer_num_pointer)
    ld (ix), a             ;; Save de counter on memory

    ;;Decrease end pointer of the array
    ld ix, (ead_pointer_end_pointer)    ;; Load end pointer pointer to hl :( xd
    ld h, 1(ix)                         ;; Load end pointer to hl
    ld l, 0(ix)
    or a                                ;; Reset the c flag so it wont be substracted in sbc
    ld bc, #2
    sbc hl, bc                          ;; Substract the size of a pointer wich is 2
    ld 1(ix), h                          ;; Save the value on memory
    ld 0(ix), l

    pop hl
    pop ix

    jp _ead_pointer_deleted             ;; Do not increase the array pointer

_ead_not_equals:
    inc ix
    inc ix                              ;; IX now points to the next pointer
_ead_pointer_deleted:
    ld a, (ead_local_counter)
    dec a
    jp _ead_loop

_end_ead_loop:
    ret